from typing import Annotated, Callable

from mongorepo._base import Access
from mongorepo._handlers import _handle_mongo_repository


def mongo_repository(
    cls: type | None = None,
    add: bool = True,
    add_batch: bool = True,
    get: bool = True,
    get_all: bool = True,
    get_list: bool = True,
    update: bool = True,
    delete: bool = True,
    integer_fields: list[Annotated[str, 'field names']] | None = None,
    array_fields: list[Annotated[str, 'field names']] | None = None,
    method_access: Access | None = None,
    __methods__: bool = False,
) -> type | Callable:
    """## Decorator for MongoDB repositories

    * decorated class must provide `Meta` class inside
    with variables "dto"(represent simple dataclass) and
    `collection` (represent mongo collection of type `Collection` from `pymongo` library)

    * You can also provide `index` field that can be just a string name of the field
    which you want to make index field or it can be instance of `mongorepo.Index`
    with extended settings

    * You can use `method_access` to make all methods
    private, protected or public (use `mongorepo.Access`)

    * Add `array_fields` to params to extend repository with methods
    that related to list fields in your dto type

    {field}__pop, {field}__list, {field}__remove, {field__append}

    * Add `integer_fields` to params to extend repository with methods
    that related to integer fields in your dto type

    increment_{field}, decrement_{field}

    * __methods__: if `True` set class property `__methods__`
    that list all available methods including `mongorepo` methods, can be useful
    if you want to debug or see the signature of generated methods

    ### Example of __methods__ property
    ```
    @mongorepo_repository(get_list=True, update=True, __methods__=True)
    class A:
        ...
    print(A().__methods__)
    # get_list(self, offset: int = 0, limit: int = 20) -> list[~DTO]
    # update(self, dto: ~DTO, **filters: Any) -> Optional[~DTO]
    ```

    ## Decorator usage example

    ```
    @mongo_repository(delete=False)
    class MongoRepository:
        class Meta:
            dto = UserDTO
            collection: Collection = db["users"]
            index = mongorepo.Index(field="name")
            method_access = mongorepo.Access.PUBLIC

    r = MongoRepository()

    r.add(UserDTO(username="admin"))

    admin = r.get(username="admin")  # UserDTO(username="admin")

    ```

    """

    def wrapper(cls) -> type:
        return _handle_mongo_repository(
            cls=cls,
            add=add,
            update=update,
            get_all=get_all,
            get=get,
            get_list=get_list,
            delete=delete,
            add_batch=add_batch,
            integer_fields=integer_fields,
            array_fields=array_fields,
            method_access=method_access,
            __methods__=__methods__,
        )

    if cls is None:
        return wrapper

    return wrapper(cls)



from dataclasses import Field, dataclass
from enum import Enum, StrEnum
from typing import Any, ClassVar, Literal, Protocol, TypeVar

from motor.motor_asyncio import AsyncIOMotorCollection
from pymongo.collection import Collection


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


DTO = TypeVar('DTO', bound=DataclassInstance)
_DTOField = TypeVar('_DTOField', bound=DataclassInstance)


@dataclass(repr=False, slots=True, eq=False)
class Index:
    """Allows to specify extra information about mongodb index."""
    field: str
    name: str | None = None
    desc: bool = True
    unique: bool = False


class Access(int, Enum):
    """Use to indicate method access in a repository."""
    PUBLIC = 0
    PROTECTED = 1
    PRIVATE = 2


@dataclass(eq=False)
class MethodDeps:
    """DTO for mongorepo methods dependencies.

    ### includes:
    ```
    collection: Collection | AsyncIOMotorCollection | None = None
    dto: type[DTO] | None = None
    id_field: str | None = None
    field_name: str | None = None
    ```

    """
    collection: Collection | AsyncIOMotorCollection
    dto_type: type
    id_field: str | None = None
    custom_field_method_name: str | None = None
    update_integer_weight: int | None = None


class MethodAction(StrEnum):
    GET = 'get'
    GET_LIST = 'get_list'
    GET_ALL = 'get_all'
    UPDATE = 'update'
    ADD = 'add'
    ADD_BATCH = 'add_batch'
    DELETE = 'delete'

    INTEGER_INCREMENT = 'incr__'
    INTEGER_DECREMENT = 'decr__'

    LIST_APPEND = '__append'
    LIST_REMOVE = '__remove'
    LIST_POP = '__pop'
    LIST_FIELD_VALUES = '__list'


class ParameterEnum(StrEnum):
    FILTER = 'filters'
    OFFSET = 'offset'
    LIMIT = 'limit'
    DTO = 'dto'
    VALUE = 'value'
    WEIGHT = 'weight'
    FILTER_ALIAS = '__filter_alias'


LParameter = Literal[
    ParameterEnum.FILTER,
    ParameterEnum.OFFSET,
    ParameterEnum.LIMIT,
    ParameterEnum.DTO,
    ParameterEnum.VALUE,
    ParameterEnum.WEIGHT,
    ParameterEnum.FILTER_ALIAS,
]



import pymongo
from motor.motor_asyncio import AsyncIOMotorCollection

from mongorepo._base import Index


async def _create_index_async(index: Index | str, collection: AsyncIOMotorCollection) -> None:
    """### Creates an index for the collection

    * index parameter can be string or mongorepo.Index
    * If index is string, create standard mongodb index
    * If it's `mongorepo.Index` creates index with user's settings

    """
    if isinstance(index, str):
        await collection.create_index(index)
        return
    index_name = f'index_{index.field}'
    if index.name:
        index_name = index.name
    direction = pymongo.DESCENDING if index.desc else pymongo.ASCENDING
    await collection.create_index(
        [(index.field, direction)],
        name=index_name,
        unique=index.unique,
    )



from .decorator import implements
from .methods import (
    AddBatchMethod,
    AddMethod,
    DeleteMethod,
    FieldAlias,
    GetAllMethod,
    GetListMethod,
    GetMethod,
    Method,
    SpecificMethod,
    UpdateMethod,
)

__all__ = [
    'FieldAlias',
    'implements',
    'AddBatchMethod',
    'AddMethod',
    'DeleteMethod',
    'GetAllMethod',
    'GetListMethod',
    'GetMethod',
    'Method',
    'SpecificMethod',
    'UpdateMethod',
]



from typing import Callable

from mongorepo._base import MethodAction
from mongorepo._methods.arrays import (
    _get_list_of_field_values_method,
    _pop_list_method,
    _update_list_field_method,
)
from mongorepo._methods.crud import (
    _add_batch_method,
    _add_method,
    _delete_method,
    _get_all_method,
    _get_list_method,
    _get_method,
    _update_method,
)
from mongorepo._methods.integer import _update_integer_field_method

# List methods, all starts with `__`
LIST_METHODS: dict[str, Callable] = {
    MethodAction.LIST_FIELD_VALUES: _get_list_of_field_values_method,
    MethodAction.LIST_POP: _pop_list_method,
    MethodAction.LIST_APPEND: _update_list_field_method,
    MethodAction.LIST_REMOVE: _update_list_field_method,
}

# Integer methods, all ends with `__`
INTEGER_METHODS: dict[str, Callable] = {
    MethodAction.INTEGER_INCREMENT: _update_integer_field_method,
    MethodAction.INTEGER_DECREMENT: _update_integer_field_method,
}

# CRUD methods
CRUD_METHODS: dict[str, Callable] = {
    MethodAction.GET: _get_method,
    MethodAction.ADD: _add_method,
    MethodAction.UPDATE: _update_method,
    MethodAction.DELETE: _delete_method,
    MethodAction.GET_LIST: _get_list_method,
    MethodAction.GET_ALL: _get_all_method,
    MethodAction.ADD_BATCH: _add_batch_method,
}

__all__ = ['LIST_METHODS', 'CRUD_METHODS', 'INTEGER_METHODS']



import inspect
from dataclasses import is_dataclass
from inspect import Parameter, _empty
from typing import Any, Callable, TypeVar

from motor.motor_asyncio import AsyncIOMotorCollection
from pymongo.collection import Collection

from mongorepo import exceptions
from mongorepo._base import DTO, MethodAction, MethodDeps
from mongorepo._base import ParameterEnum as MongorepoParameter
from mongorepo.utils import (
    _get_defaults,
    _replace_typevars,
    _validate_method_annotations,
)

from ._utils import (
    _get_method_from_string,
    _get_mongorepo_method_callable,
    _get_validated_mongorepo_method_callable,
)
from .methods import Method, SpecificFieldMethod, SpecificMethod


def _substitute_specific_method(
    deps: MethodDeps,
    method: SpecificMethod | SpecificFieldMethod,
) -> Callable:
    is_async = inspect.iscoroutinefunction(method.source)

    if hasattr(method, 'field_name'):
        deps.custom_field_method_name = getattr(method, 'field_name')
    if hasattr(method, 'integer_weight'):
        deps.update_integer_weight = getattr(method, 'integer_weight')

    callable_mongorepo_method = _get_mongorepo_method_callable(
        method.action,
        method.mongorepo_method,
        deps,
    )

    def func(self, *args, **kwargs) -> Any:
        required_params = _manage_custom_params(
            Method(source=method.source, **method.params), *args, **kwargs,
        )
        return callable_mongorepo_method(self, **required_params)

    async def async_func(self, *args, **kwargs) -> Any:
        required_params = _manage_custom_params(
            Method(source=method.source, **method.params), *args, **kwargs,
        )
        return await callable_mongorepo_method(self, **required_params)

    if method.action == MethodAction.GET_ALL and is_async is True:
        new_method = func
    else:
        new_method = async_func if is_async else func

    new_method.__annotations__ = method.source.__annotations__
    new_method.__name__ = method.name

    _replace_typevars(new_method, deps.dto_type)

    return new_method


def _substitute_method(
    mongorepo_method_name: str,
    generic_method: Method,
    dto: type[DTO],
    collection: Collection | AsyncIOMotorCollection,
    id_field: str | None = None,
) -> Callable:

    mongorepo_method_function, mongorepo_method_type = _get_method_from_string(
        mongorepo_method_name, is_async=generic_method.is_async,
    )
    mongorepo_method = _get_validated_mongorepo_method_callable(
        mongorepo_method_name,
        mongorepo_method_type,
        mongorepo_method_function,
        dto,
        collection,
        id_field,
    )

    _validate_method_annotations(generic_method.source)

    def func(self, *args, **kwargs) -> Any:
        required_params = _manage_params(
            mongorepo_method, generic_method, *args, **kwargs,
        )
        return mongorepo_method(self, **required_params)

    async def async_func(self, *args, **kwargs) -> Any:
        required_params = _manage_params(
            mongorepo_method, generic_method, *args, **kwargs,
        )
        return await mongorepo_method(self, **required_params)

    new_method = async_func if generic_method.is_async else func

    new_method.__annotations__ = generic_method.source.__annotations__
    new_method.__name__ = generic_method.name

    _replace_typevars(new_method, dto)

    return new_method


def _manage_custom_params(
    method: Method,
    *args,
    **kwargs,
) -> dict[str, Any]:
    defaults = _get_defaults(method.source)
    result: dict[str, Any] = {}
    filters: dict[str, Any] = {}
    source_params_map: dict[str, Any] = {}
    aliases: dict[str, Any] = method.params.get(MongorepoParameter.FILTER_ALIAS, {})  # type: ignore
    i = 0

    # Iterate over all names of the method's arguments
    for source_param in method.get_source_params().keys():

        # Parameter was passed as positional argument
        if len(args) > i:
            source_params_map[source_param] = args[i]
            i += 1
            continue

        # Parameter was passed as keyword argument ?
        elif source_param in kwargs:
            source_params_map[source_param] = kwargs[source_param]

        # If parameter was not passed check for defaults
        elif defaults.get(source_param, None):
            source_params_map[source_param] = defaults[source_param]

        # Missing parameter
        else:
            raise exceptions.MongoRepoException(
                message=f'Cannot find value for {source_param} parameter. '
                f'{method.name}() parameters: {method.params}',
            )

    for key, value in source_params_map.items():
        if method.params.get(key, None) == MongorepoParameter.FILTER:
            filters[key] = value
        elif (dto_field := aliases.get(key, None)) is not None:
            filters[dto_field] = value
        else:
            result[method.params[key]] = value

    result.update(filters)
    return result


def _manage_params(
    mongorepo_method: Callable,
    generic_method: Method,
    *args,
    **kwargs,
) -> dict[str, Any]:
    """Return required params for `mongorepo` method, do not pass `self` in
    *args or **kwargs."""

    if generic_method.params:
        return _manage_custom_params(
            generic_method, *args, **kwargs,
        )

    params: dict[str, Any] = {}
    result: dict[str, Any] = {}

    # control for **kwargs in source method
    extra: bool = False

    mongo_params = dict(inspect.signature(mongorepo_method).parameters)
    mongo_params.pop('self')

    gen_params = dict(generic_method.signature.parameters)
    gen_params.pop('self')

    i = 0
    # Looping over *args and unite them with generic method parameters
    for param in gen_params.values():
        try:
            arg = args[i]
        except IndexError:
            break
        i += 1
        params[param.name] = arg

    # Looping over **kwargs and unite them with generic method parameters
    for param in gen_params.values():
        if param.name in kwargs and param.name in params:
            raise TypeError(
                f'{generic_method.name}() keyword arguments repeated: {param.name}',
            )

        if param.default != _empty and not kwargs.get(param.name, None):
            params[param.name] = param.default

        if param.name in params:
            continue

        if param.name not in params and param.name not in kwargs:
            raise TypeError(
                f'{generic_method.name}() '
                f'missing required keyword argument: \'{param.name}\'',
            )
        params[param.name] = kwargs[param.name]

    # Unite generic method arguments to mongorepo arguments of the method
    for mongo_param in mongo_params.values():
        if mongo_param.kind is Parameter.VAR_KEYWORD:
            extra = True
            break

        # Try to find DTO type and unite with parameter with DTO annotation
        if isinstance(mongo_param.annotation, TypeVar) or is_dataclass(mongo_param.annotation):
            for key, value in params.items():
                if is_dataclass(value):
                    result[mongo_param.name] = value
                    params[key] = None
                    break
        else:
            for key, value in params.items():
                if mongo_param.annotation == Any or isinstance(value, mongo_param.annotation):
                    result[mongo_param.name] = value
                    params[key] = None
                    break

    # Looping over mongorepo **kwargs if they not absent
    if extra:
        for key, value in params.items():
            if value is not None:
                result[key] = value
    return result



from dataclasses import asdict
from typing import Any, Generic, Iterable

from pymongo.collection import Collection

from mongorepo import exceptions
from mongorepo._base import DTO, Index
from mongorepo._methods import _add_method
from mongorepo.utils import (
    _create_index,
    _get_converter,
    _get_dto_from_origin,
    _get_meta_attributes,
)


class BaseMongoRepository(Generic[DTO]):
    """
    ## Base MongoDB repository class
    #### Extends child classes with various methods:

    ```
    add(self, dto: DTO) -> DTO
    get(self, **filters) -> DTO | None
    get_all(self, **filters) -> Iterable[DTO]
    update(self, dto: DTO, **filters) -> DTO
    delete(self, **filters) -> bool
    ```

    #### Provide DTO type in type hints, example:

    ```
    class DummyMongoRepository(BaseMongoRepository[UserDTO]):
        ...
    ```

    * If you want to create an index use `mongorepo.Index`
      or just a name of the field to put index on
    """

    def __new__(cls, *args, **kwargs) -> 'BaseMongoRepository':
        instance = super().__new__(cls)
        dto_type = _get_dto_from_origin(cls)
        setattr(instance, 'dto_type', dto_type)

        try:
            meta = _get_meta_attributes(cls, raise_exceptions=False)
        except exceptions.NoMetaException:
            meta = None

        index: Index | str | None = meta['index'] if meta else None
        id_field: str | None = meta['id_field'] if meta else None

        setattr(instance, '__convert_to_dto', _get_converter(dto_type, id_field=id_field))
        setattr(instance, '__id_field', id_field)

        if index is not None:
            collection: Collection | None = meta['collection'] if meta else None
            if collection is None:
                raise exceptions.NoCollectionException(
                    message='Cannot access collection from Meta to create index',
                )
            _create_index(index, collection=collection)
        return instance

    def __init__(self, collection: Collection | None = None) -> None:
        self.collection = self.__get_collection(collection)
        self.__convert_to_dto = self.__dict__['__convert_to_dto']
        self.dto_type = self.__dict__.get('dto_type', None) or _get_dto_from_origin(self.__class__)
        self.__add = _add_method(
            dto_type=self.dto_type,
            collection=self.collection,
            id_field=self.__dict__['__id_field'],
        )

    @classmethod
    def __get_collection(cls, collection: Collection | None) -> Collection:
        if collection is not None:
            return collection
        try:
            meta = _get_meta_attributes(cls, raise_exceptions=False)
        except exceptions.NoMetaException:
            raise exceptions.MongoRepoException(
                message='"Meta" class with "collecton" was not defined in the class',
            )
        if meta['collection'] is None:
            raise exceptions.NoCollectionException
        defined_collection = meta['collection']
        return defined_collection

    def _convert_to_dto(self, dct: dict[str, Any]) -> DTO:
        return self.__convert_to_dto(self.dto_type, dct)

    def get(self, **filters: Any) -> DTO | None:
        result = self.collection.find_one(filters)
        return self._convert_to_dto(result) if result else None

    def get_all(self, **filters: Any) -> Iterable[DTO]:
        cursor = self.collection.find(filters)
        for doc in cursor:
            yield self._convert_to_dto(doc)

    def get_list(self, offset: int = 0, limit: int = 20) -> list[DTO]:
        cursor = self.collection.find().skip(offset).limit(limit)
        return [self._convert_to_dto(doc) for doc in cursor]

    def update(self, dto: DTO, **filters: Any) -> DTO:
        data: dict[str, dict[str, Any]] = {'$set': {}}
        for field, value in asdict(dto).items():
            data['$set'][field] = value
        self.collection.find_one_and_update(filter=filters, update=data)
        return dto

    def delete(self, **filters: Any) -> bool:
        """Returns True if document was deleted else False."""
        deleted = self.collection.find_one_and_delete(filters)
        return True if deleted else False

    def add(self, dto: DTO) -> DTO:
        return self.__add(self, dto)



from typing import Callable

from mongorepo._base import MethodAction
from mongorepo.asyncio._methods.arrays import (
    _get_list_of_field_values_method_async,
    _pop_list_method_async,
    _update_list_field_method_async,
)
from mongorepo.asyncio._methods.crud import (
    _add_batch_method_async,
    _add_method_async,
    _delete_method_async,
    _get_all_method_async,
    _get_list_method_async,
    _get_method_async,
    _update_method_async,
)
from mongorepo.asyncio._methods.integer import (
    _update_integer_field_method_async,
)

# List methods, all starts with `__`
LIST_METHODS_ASYNC: dict[str, Callable] = {
    MethodAction.LIST_FIELD_VALUES: _get_list_of_field_values_method_async,
    MethodAction.LIST_POP: _pop_list_method_async,
    MethodAction.LIST_APPEND: _update_list_field_method_async,
    MethodAction.LIST_REMOVE: _update_list_field_method_async,
}

# Integer methods, all ends with `__`
INTEGER_METHODS_ASYNC: dict[str, Callable] = {
    MethodAction.INTEGER_INCREMENT: _update_integer_field_method_async,
    MethodAction.INTEGER_DECREMENT: _update_integer_field_method_async,
}

# CRUD methods
CRUD_METHODS_ASYNC: dict[str, Callable] = {
    MethodAction.GET: _get_method_async,
    MethodAction.ADD: _add_method_async,
    MethodAction.UPDATE: _update_method_async,
    MethodAction.DELETE: _delete_method_async,
    MethodAction.GET_LIST: _get_list_method_async,
    MethodAction.GET_ALL: _get_all_method_async,
    MethodAction.ADD_BATCH: _add_batch_method_async,
}

__all__ = ['LIST_METHODS_ASYNC', 'CRUD_METHODS_ASYNC', 'INTEGER_METHODS_ASYNC']



import inspect
from typing import Callable

from mongorepo import exceptions
from mongorepo._base import MethodDeps
from mongorepo.utils import _get_meta_attributes, raise_exc

from ._substitute import _substitute_method, _substitute_specific_method
from .methods import Method, SpecificMethod


def _handle_implements_custom_methods(
    base_cls: type,
    cls: type,
    **substitute: str | Callable | Method,
) -> type:
    attrs = _get_meta_attributes(cls)
    if not substitute:
        substitute = attrs['substitute'] if attrs['substitute'] is not None else raise_exc(
            exceptions.MongoRepoException(message='No "substitute" in Meta class'),
        )
    dto_type = attrs['dto']
    collection = attrs['collection']
    id_field = attrs['id_field']
    for mongorepo_method_name, _generic_method in substitute.items():
        if not isinstance(_generic_method, Method):
            if inspect.isfunction(_generic_method) or inspect.ismethod(_generic_method):
                generic_method = Method(_generic_method)

            elif isinstance(_generic_method, str):
                __generic_method: Callable | None = getattr(
                    base_cls, _generic_method, None,
                )
                if __generic_method is None or not inspect.isfunction(__generic_method):
                    raise exceptions.InvalidMethodNameException(_generic_method)
                generic_method = Method(__generic_method)  # type: ignore
        else:
            generic_method = _generic_method

        setattr(
            cls,
            generic_method.name,
            _substitute_method(
                mongorepo_method_name,
                generic_method,
                dto_type,
                collection,
                id_field=id_field,
            ),
        )

    return cls


def _handle_implements_specific_methods(
    cls: type,
    *specific_methods: SpecificMethod,
) -> type:
    attrs = _get_meta_attributes(cls)
    dto_type = attrs['dto']
    collection = attrs['collection']
    id_field = attrs['id_field']

    for method in specific_methods:
        setattr(
            cls,
            method.name,
            _substitute_specific_method(
                MethodDeps(collection=collection, id_field=id_field, dto_type=dto_type),
                method=method,
            ),
        )
    return cls



"""Library for generating repositories for `MongoDB`,
* public API available in this module for example:
```
@mongorepo.repository
class A:
    ...
```
### Or
use `mongorepo.decorators` or `mongorepo.classes`.

Lib also has async support, so
you can use the same classes and decorators located in `mongorepo.asyncio`

"""
from . import asyncio, decorators, exceptions
from ._base import DTO, Access, Index
from .asyncio.classes import AsyncBasedMongoRepository
from .asyncio.decorators import async_mongo_repository as async_repository
from .classes import BaseMongoRepository
from .decorators import mongo_repository as repository
from .implements.decorator import implements
from .implements.methods import Method
from .queries import AggregationStage, Condition, Operation, UpdateModifier

__all__ = [
    'Method',
    'UpdateModifier',
    'AggregationStage',
    'Operation',
    'Condition',
    'implements',
    'BaseMongoRepository',
    'AsyncBasedMongoRepository',
    'repository',
    'async_repository',
    'Access',
    'DTO',
    'Index',
    'asyncio',
    'decorators',
    'exceptions',
]



from enum import Enum
from typing import Callable

from motor.motor_asyncio import AsyncIOMotorCollection
from pymongo.collection import Collection

from mongorepo import exceptions
from mongorepo._base import DTO, MethodAction, MethodDeps
from mongorepo._methods import CRUD_METHODS, INTEGER_METHODS, LIST_METHODS
from mongorepo.asyncio._methods import (
    CRUD_METHODS_ASYNC,
    INTEGER_METHODS_ASYNC,
    LIST_METHODS_ASYNC,
)
from mongorepo.utils import _check_valid_field_type


class MethodType(Enum):
    CRUD = 1
    LIST = 2
    INTEGER = 3

    @staticmethod
    def from_action(action: MethodAction) -> 'MethodType':
        if action.value.startswith('__'):
            return MethodType.LIST
        elif action.value.endswith('__'):
            return MethodType.INTEGER
        elif action in MethodAction:
            return MethodType.CRUD
        raise exceptions.InvalidActionException(
            action=action, valid_actions=[a.value for a in MethodAction],
        )


def _get_method_from_string(
    method_name: str, is_async: bool = False,
) -> tuple[Callable, MethodType]:
    """Tries to get right `mongorepo` method according to `method_name`

    returns tuple where first element is a method, and the second its type

    ### Example:

    ```
    method = get_method_from_string(method_name='messages__list', is_async=True)
    ```

    """
    if is_async:
        if method_name in CRUD_METHODS_ASYNC:
            return (CRUD_METHODS_ASYNC[method_name], MethodType.CRUD)

        for key, value in LIST_METHODS_ASYNC.items():
            if method_name.endswith(key):
                return (value, MethodType.LIST)

        for key, value in INTEGER_METHODS_ASYNC.items():
            if method_name.startswith(key):
                return (value, MethodType.INTEGER)
    else:
        if method_name in CRUD_METHODS:
            return (CRUD_METHODS[method_name], MethodType.CRUD)

        for key, value in LIST_METHODS.items():
            if method_name.endswith(key):
                return (value, MethodType.LIST)

        for key, value in INTEGER_METHODS.items():
            if method_name.startswith(key):
                return (value, MethodType.INTEGER)

    msg = (
        f'Mongorepo does not have method with this action: {method_name}\n'
        f'Use valid prefix or suffix to specify correct method that you want to use\n'
        f'For example: "records__pop" where "records" is the name of list field and '
        '"__pop" action to pop the last element in the list'
    )

    raise exceptions.InvalidMethodNameException(
        method_name=method_name,
        message=msg,
    )


# TODO: reduce repeated parts with _get_mongorepo_method_callable()
def _get_validated_mongorepo_method_callable(
    mongorepo_method_name: str,
    mongorepo_method_type: MethodType,
    mongorepo_method: Callable,
    dto: type[DTO],
    collection: Collection | AsyncIOMotorCollection,
    id_field: str | None = None,
) -> Callable:
    """Tries to parse `mongorepo_method_name` to get method action.

    After parsing passes required parameters to the mongorepo method
    decorator to get callable method

    """
    if mongorepo_method_type is MethodType.CRUD:
        if id_field in mongorepo_method.__annotations__:
            mongorepo_method = mongorepo_method(
                dto_type=dto, collection=collection, id_field=id_field,
            )
        else:
            mongorepo_method = mongorepo_method(dto_type=dto, collection=collection)

    elif mongorepo_method_type is MethodType.INTEGER:
        action, field_name = (d := mongorepo_method_name.split('__'))[0] + '__', d[-1]
        _check_valid_field_type(field_name, dto, int)
        if action not in MethodAction:
            raise exceptions.InvalidActionException(
                action=action,
                valid_actions=(valid_actions := [act.value for act in MethodAction]),
                message=f'Invalid action for {mongorepo_method_name}(): '
                f'{action}\n valid: {valid_actions}',
            )
        mongorepo_method = mongorepo_method(
            dto_type=dto,
            collection=collection,
            field_name=field_name,
            _weight=1 if action == MethodAction.INTEGER_INCREMENT else -1,
        )

    elif mongorepo_method_type is MethodType.LIST:
        action, field_name = '__' + (d := mongorepo_method_name.split('__'))[-1], d[0]
        _check_valid_field_type(field_name, dto, list)
        if action not in MethodAction:
            raise exceptions.InvalidActionException(
                action=action,
                valid_actions=(valid_actions := [act.value for act in MethodAction]),
                message=f'Invalid action for {mongorepo_method_name}(): '
                f'{action}\n valid: {valid_actions}',
            )
        if 'command' in mongorepo_method.__annotations__:
            mongorepo_method = mongorepo_method(
                dto_type=dto,
                collection=collection,
                field_name=field_name,
                command='$push' if action == MethodAction.LIST_APPEND else '$pull',
            )
        else:
            mongorepo_method = mongorepo_method(
                dto_type=dto,
                collection=collection,
                field_name=field_name,
            )
    return mongorepo_method


def _get_mongorepo_method_callable(
    action: MethodAction,
    mongorepo_method: Callable,
    deps: MethodDeps,
) -> Callable:
    """Tries to parse `mongorepo_method_name` to get method action.

    After parsing passes required parameters to the mongorepo method
    decorator to get callable method

    """
    mongorepo_method_type = MethodType.from_action(action)

    match mongorepo_method_type:
        case MethodType.CRUD:
            if deps.id_field in mongorepo_method.__annotations__:
                mongorepo_method = mongorepo_method(
                    dto_type=deps.dto_type, collection=deps.collection, id_field=deps.id_field,
                )
            else:
                mongorepo_method = mongorepo_method(
                    dto_type=deps.dto_type, collection=deps.collection,
                )
        case MethodType.INTEGER:
            mongorepo_method = mongorepo_method(
                dto_type=deps.dto_type,
                collection=deps.collection,
                field_name=deps.custom_field_method_name,
                #      increment                                           decrement
                _weight=1 if action == MethodAction.INTEGER_INCREMENT else -1,
            )
        case MethodType.LIST:
            if 'command' in mongorepo_method.__annotations__:
                mongorepo_method = mongorepo_method(
                    dto_type=deps.dto_type,
                    collection=deps.collection,
                    field_name=deps.custom_field_method_name,
                    #       append                                             remove
                    command='$push' if action == MethodAction.LIST_APPEND else '$pull',
                )
            else:
                # 'pop' method or 'list of field values' method
                mongorepo_method = mongorepo_method(
                    dto_type=deps.dto_type,
                    collection=deps.collection,
                    field_name=deps.custom_field_method_name,
                )
    return mongorepo_method



from dataclasses import dataclass, field


@dataclass(eq=False)
class MongoRepoException(Exception):
    message: str | None = field(kw_only=True, default=None)


@dataclass(eq=False)
class NoMetaException(MongoRepoException):

    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'Class does not have "Meta" class inside'


@dataclass(eq=False)
class NoDTOTypeException(MongoRepoException):
    with_meta: bool = True

    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'DTO type was not provided' + ' in "Meta" class' if self.with_meta else ''


@dataclass(eq=False)
class NoCollectionException(MongoRepoException):
    with_meta: bool = True

    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'Collection was not provided' + ' in "Meta" class' if self.with_meta else ''


class NotDataClass(MongoRepoException):
    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'Provided dto type does not implement dataclass interface'


@dataclass(eq=False)
class InvalidMethodNameException(MongoRepoException):
    method_name: str
    available_methods: tuple[str, ...] | None = None

    def __str__(self) -> str:
        if self.message:
            return self.message
        message = f'Method "{self.method_name}" does not appear in mongorepo available methods'
        if self.available_methods:
            message += f'\nAvailable methods: {self.available_methods!r}'
        return message


@dataclass(eq=False)
class TypeHintException(MongoRepoException):
    def __str__(self) -> str:
        if self.message:
            return self.message
        message = 'Invalid type hint'
        return message


class NotFoundException(MongoRepoException):
    def __init__(self, **filters) -> None:
        self.filters = filters

    def __str__(self) -> str:
        filters = ', '.join([f'{key}={value}' for key, value in self.filters.items()])
        return f'Document not found, filters: {filters}'


@dataclass(eq=False)
class InvalidActionException(MongoRepoException):
    action: str
    valid_actions: list[str]

    def __str__(self) -> str:
        return (
            self.message or f'Invalid method action: {self.action}\nValid are: {self.valid_actions}'
        )



from typing import Callable, overload

from ._handlers import (
    _handle_implements_custom_methods,
    _handle_implements_specific_methods,
)
from .methods import Method, SpecificMethod


@overload
def implements(base_cls: type, **custom_methods: str | Callable | Method) -> Callable:
    ...


@overload
def implements(base_cls: type) -> Callable:
    ...


@overload
def implements(base_cls: type, *specific_methods: SpecificMethod) -> Callable:
    ...


def implements(
    base_cls,
    *specific_methods: SpecificMethod,
    **custom_methods: str | Method | Callable,
) -> Callable:
    """Decorator that allows to implement methods of `base_cls`

    ## [Not recommended] Simple Usage Example:

    ```
    class A:
        def get_my_entity(self, id: str):
            ...

    @implements(A, get=A.get_my_entity)
    class MongoRepo:
        class Meta:
            dto = SimpleDTO
            collection = random_collection()

    my_entity = MongoRepo().get_my_entity(id="10")

    ```
    ## [Recommended] Usage with `SpecificMethod` protocol example:

    ```
    class IRepo:
        async def get(self, title: str) -> SomeDataclass:
            ...

        async def add(self, model: SomeDataclass) -> SomeDataclass:
            ...

    @implements(
        IRepo,
        GetMethod(IRepo.get, filters=['title']),
        AddMethod(IRepo.add, dto='model'),
    )
    class MongoRepo:
        class Meta:
            collection = some_collection()
            dto = SomeDataclass

    r: IRepo = MongoRepo()
    await r.add(SomeDataclass('some title'))
    dto = await r.get(title='some title')
    print(dto)  # SomeDataclass(title='some title')
    ```

    """
    if specific_methods and custom_methods:
        raise ValueError(
            'Specify either positional arguments (*specific_methods)'
            'or keyword arguments (**custom_methods), not both.',
        )

    def wrapper(cls) -> type:
        if custom_methods:
            return _handle_implements_custom_methods(
                base_cls=base_cls,
                cls=cls,
                **custom_methods,
            )
        elif specific_methods:
            return _handle_implements_specific_methods(
                cls,
                *specific_methods,
            )
        else:
            return _handle_implements_custom_methods(
                base_cls=base_cls,
                cls=cls,
                **custom_methods,
            )
    return wrapper



def get_available_meta_attributes(only_names: bool = False) -> dict[str, str] | tuple[str, ...]:
    """Returns dictionary of available Meta attributes, Meta is a class that
    contains meta data for `mongorepo` decorators and classes.

    Use `only_names=True` to get names of this attributes

    """
    attrs = {
        'index': 'creates index for a collection',
        'method_access': (
            'added methods will be private, '
            'protected or public, use mongorepo.Access'
        ),
        'dto': 'sets default dto for repository, repository saves data in the format of the dto',
        'collection': 'sets default collection for repository',
        'id_field': (
            'mongorepo will store MongoDB `_id` there in string format, '
            'only define name of the DTO field in string format'
        ),
        'substitute': (
            'dict[str, str], use for mongorepo.decorators.implements decorator '
            'can be replaced with **methods params in the @implements'
        ),
    }
    if only_names:
        return tuple(attrs.keys())
    return attrs


def get_available_repository_methods(only_names: bool = False) -> dict[str, str] | tuple[str, ...]:
    """Returns dictionary of available methods for `mongorepo` decorators or
    classes.

    Use `only_names=True` to get names of this methods

    """
    methods = {
        'add': 'add document to a collection, collection stores it in dto format, params: dto: DTO',
        'delete': 'delete document from a collection, params: **filters: Any',
        'update': (
            'update document in a collection takes dto as parameter, '
            'if no value in dto(except: float, int, bool) skips it'
        ),
        'get': 'retrieve a document from a collection, params: **filters: Any',
        'get_all': 'retrive all documents of the same type(dto type), params: **filters: Any',
        'get_list': 'collects bunch of documents, params: offset: int = 0, limit: int = 20',
        '{field}__append': (
            'Only for decorator with "array_fields", append item to a document'
        ),
        '{field}__remove': (
            'Only for decorator with "array_fields", remove item from a document'
        ),
        '{field}__pop': (
            'Only for decorator with "array_fields", pop last item from a document'
        ),
        '{field}__list': (
            'Only for decorator with "array_fields", return list of values of field'
        ),
        'incr__{field}': (
            'Only for decorator with "integer_fields", increment integer field by 1 in a document'
        ),
        'decr__{field}': (
            'Only for decorator with "integer_fields", decrement integer fied by 1 in a document'
        ),
    }
    if only_names:
        return tuple(methods.keys())
    return methods



import asyncio

from mongorepo._base import Access
from mongorepo._setters import (
    _set_array_fields_methods,
    _set_crud_methods,
    _set_integer_fields_methods,
)
from mongorepo.asyncio.utils import _create_index_async
from mongorepo.utils import (
    _create_index,
    _get_meta_attributes,
    _set__methods__,
)


def _handle_mongo_repository(
    cls,
    add: bool,
    get: bool,
    add_batch: bool,
    get_all: bool,
    update: bool,
    delete: bool,
    get_list: bool,
    __methods__: bool,
    integer_fields: list[str] | None,
    array_fields: list[str] | None,
    method_access: Access | None,
) -> type:
    """Calls for functions that set different methods and attributes to the
    class."""
    attributes = _get_meta_attributes(cls, raise_exceptions=False)
    collection = attributes['collection']
    index = attributes['index']

    _set_crud_methods(
        cls,
        add=add,
        get=get,
        get_all=get_all,
        get_list=get_list,
        update=update,
        delete=delete,
        add_batch=add_batch,
        method_access=method_access,
    )

    if index is not None:
        _create_index(index=index, collection=collection)

    if integer_fields is not None:
        _set_integer_fields_methods(cls, integer_fields=integer_fields, method_access=method_access)

    if array_fields is not None:
        _set_array_fields_methods(cls, array_fields=array_fields, method_access=method_access)

    if __methods__:
        _set__methods__(cls)
    return cls


def _handle_async_mongo_repository(
    cls,
    add: bool,
    add_batch: bool,
    get: bool,
    get_all: bool,
    get_list: bool,
    update: bool,
    delete: bool,
    __methods__: bool,
    integer_fields: list[str] | None,
    array_fields: list[str] | None,
    method_access: Access | None,
) -> type:
    """Calls for functions that set different async methods and attributes to
    the class."""
    attributes = _get_meta_attributes(cls, raise_exceptions=False)
    collection = attributes['collection']
    index = attributes['index']

    _set_crud_methods(
        cls,
        add=add,
        get=get,
        get_all=get_all,
        update=update,
        delete=delete,
        add_batch=add_batch,
        method_access=method_access,
        get_list=get_list,
        async_methods=True,
    )

    if integer_fields is not None:
        _set_integer_fields_methods(
            cls, integer_fields=integer_fields, async_methods=True, method_access=method_access,
        )

    if array_fields is not None:
        _set_array_fields_methods(
            cls, array_fields=array_fields, async_methods=True, method_access=method_access,
        )

    if index is not None:
        asyncio.create_task(_create_index_async(index=index, collection=collection))

    if __methods__:
        _set__methods__(cls)
    return cls



from dataclasses import asdict
from typing import Any, AsyncGenerator, Callable

from bson import ObjectId
from motor.motor_asyncio import AsyncIOMotorCollection

from mongorepo import exceptions
from mongorepo._base import DTO
from mongorepo.utils import _get_converter, raise_exc


def _add_method_async(
    dto_type: type[DTO],
    collection: AsyncIOMotorCollection,
    id_field: str | None = None,
) -> Callable:

    async def add(self, dto: DTO) -> DTO:
        await collection.insert_one(asdict(dto))
        return dto

    async def add_and_return_with_id(self, dto: DTO) -> DTO:
        object_id = ObjectId()
        dto.__dict__[id_field] = str(object_id)  # type: ignore
        await collection.insert_one({**asdict(dto), '_id': object_id})
        return dto

    if not id_field:
        return add
    return add_and_return_with_id


def _add_batch_method_async(
    dto_type: type[DTO],
    collection: AsyncIOMotorCollection,
    id_field: str | None = None,
) -> Callable:

    async def add(self, dto_list: list[DTO]) -> None:
        await collection.insert_many(asdict(d) for d in dto_list)

    async def add_and_return_with_id(self, dto_list: list[DTO]) -> None:
        batch: list[dict[str, Any]] = []
        for dto in dto_list:
            object_id = ObjectId()
            dto.__dict__[id_field] = str(object_id)  # type: ignore
            batch.append({**asdict(dto), '_id': object_id})
        await collection.insert_many(batch)

    if id_field is not None:
        return add
    return add_and_return_with_id


def _get_list_method_async(
    dto_type: type[DTO],
    collection: AsyncIOMotorCollection,
    id_field: str | None = None,
):
    to_dto = _get_converter(dto_type, id_field=id_field)

    async def get_list(self, offset: int = 0, limit: int = 20, **filters) -> list[DTO]:
        cursor = collection.find(filter=filters).skip(offset).limit(limit)
        return [to_dto(dto_type, doc) async for doc in cursor]
    return get_list


def _get_all_method_async(
    dto_type: type[DTO],
    collection: AsyncIOMotorCollection,
    id_field: str | None = None,
) -> Callable:
    to_dto = _get_converter(dto_type, id_field=id_field)

    async def get_all(self, **filters: Any) -> AsyncGenerator[DTO, None]:
        cursor = collection.find(filters)
        async for dct in cursor:
            yield to_dto(dto_type, dct)
    return get_all


def _update_method_async(
    dto_type: type[DTO],
    collection: AsyncIOMotorCollection,
    id_field: str | None = None,
) -> Callable:
    to_dto = _get_converter(dto_type, id_field=id_field)

    async def update(self, dto: DTO, **filters: Any) -> DTO | None:
        data: dict[str, dict[str, Any]] = {'$set': {}}
        for field, value in asdict(dto).items():
            data['$set'][field] = value
        doc: dict | None = await collection.find_one_and_update(
            filter=filters, update=data, return_document=True,
        )
        return to_dto(dto_type, doc) if doc else None
    return update


def _delete_method_async(dto_type: type[DTO], collection: AsyncIOMotorCollection) -> Callable:
    async def delete(self, **filters: Any) -> bool:
        deleted = await collection.find_one_and_delete(filters)
        return True if deleted else False
    return delete


def _get_method_async(
    dto_type: type[DTO],
    collection: AsyncIOMotorCollection,
    id_field: str | None = None,
) -> Callable:
    to_dto = _get_converter(dto_type, id_field=id_field)

    async def get(self, _id: str | None = None, **filters: Any) -> DTO | None:
        result = await collection.find_one(filters)
        return to_dto(dto_type, result) if result else None
    return get


def _update_field_method_async(
    dto_type: type[DTO],
    collection: AsyncIOMotorCollection,
    id_field: str | None = None,
) -> Callable:
    to_dto = _get_converter(dto_type, id_field=id_field)

    async def update_field(self, field_name: str, value: Any, **filters) -> DTO | None:
        if field_name not in dto_type.__dict__['__annotations__']:
            raise exceptions.MongoRepoException(
                message=f'{dto_type} does not have field "{field_name}"',
            )
        document = await collection.find_one_and_update(
            filter=filters, update={'$set': {field_name: value}}, return_document=True,
        )
        raise_exc(exceptions.NotFoundException(**filters)) if not document else ...
        return to_dto(dto_type, document)
    return update_field



from typing import Callable

from pymongo.collection import Collection

from mongorepo import exceptions
from mongorepo._base import DTO
from mongorepo.utils import raise_exc


def _update_integer_field_method(
    dto_type: type[DTO], collection: Collection, field_name: str, _weight: int = 1,
) -> Callable:
    def update_integer_field(self, weight: int | None = None, **filters) -> None:
        w = weight if weight is not None else _weight
        doc = collection.update_one(
            filter=filters, update={'$inc': {field_name: w}},
        )
        raise_exc(exceptions.NotFoundException(**filters)) if not doc else ...
    return update_integer_field
